"""Tests for P1.1 — Rule-based traceback informativeness pre-gate."""

from __future__ import annotations

import shutil
from dataclasses import dataclass, field
from pathlib import Path

import pytest

from evals import run_eval
from evals.patchers.base import ContextRequestResult, PatchResult

# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------


def _make_case(tmp_path: Path) -> run_eval.CaseSpec:
    case_dir = tmp_path / "case_src"
    case_dir.mkdir(parents=True, exist_ok=True)
    (case_dir / "buggy.py").write_text(
        "def f() -> int:\n    return 1\n",
        encoding="utf-8",
    )
    (case_dir / "test_buggy.py").write_text(
        "from buggy import f\n\ndef test_f() -> None:\n    assert f() == 2\n",
        encoding="utf-8",
    )
    return run_eval.CaseSpec(
        case_id="pregate_case",
        path=case_dir,
        category="traceback_controls",
        difficulty="easy",
        snapshot_dependency="helpful",
        expected_exception="AssertionError",
        tags=[],
        python_args=["-m", "pytest", "-q"],
        timeout_sec=30,
    )


def _snapshot_payload(
    exc_type: str = "AssertionError",
    exc_message: str = "expected 2 got 1",
    frames: list[dict] | None = None,
) -> dict:
    if frames is None:
        frames = [
            {
                "file_rel": "buggy.py",
                "line": 2,
                "function": "f",
                "code": "return 1",
                "locals": {"x": 1},
            }
        ]
    return {
        "exception": {"type": exc_type, "message": exc_message},
        "frames": frames,
        "pytest": {"nodeid": "test_buggy.py::test_f", "repro": ["python", "-m", "pytest", "-q"]},
    }


@dataclass
class _PreGateCapturePatcher:
    """Patcher that records whether request_context was called."""

    name: str = "pregate_capture"
    request_calls: int = 0
    prompts: list[str] = field(default_factory=list)

    def request_context(self, *, prompt: str, context_dir, cwd) -> ContextRequestResult:
        del context_dir, cwd
        self.request_calls += 1
        return ContextRequestResult(
            request_json={
                "need_more_context": True,
                "requests": [{"type": "file", "path": "buggy.py"}],
                "sufficiency_rationale": "need code context",
            },
            raw_response='{"need_more_context":true}',
            meta={},
        )

    def propose_patch(self, *, prompt: str, context_dir, cwd) -> PatchResult:
        del context_dir, cwd
        self.prompts.append(prompt)
        return PatchResult(diff=None, meta={"reason": "no_diff"})


# ===========================================================================
# Unit tests: Scoring
# ===========================================================================


class TestTracebackInformativenessScore:
    def test_syntax_error_scores_high(self) -> None:
        result = run_eval._traceback_informativeness_score(
            exception_type="SyntaxError",
            exception_message="invalid syntax",
            frame_count=1,
            crash_in_user_code=True,
        )
        assert result.score >= 0.85

    def test_name_error_scores_high(self) -> None:
        result = run_eval._traceback_informativeness_score(
            exception_type="NameError",
            exception_message="name 'x' is not defined",
            frame_count=1,
            crash_in_user_code=True,
        )
        assert result.score >= 0.8

    def test_indentation_error_scores_high(self) -> None:
        result = run_eval._traceback_informativeness_score(
            exception_type="IndentationError",
            exception_message="unexpected indent",
            frame_count=1,
            crash_in_user_code=True,
        )
        assert result.score >= 0.85

    def test_assertion_error_bare_scores_low(self) -> None:
        result = run_eval._traceback_informativeness_score(
            exception_type="AssertionError",
            exception_message="",
            frame_count=1,
            crash_in_user_code=True,
        )
        # 0.30*0.1 + 0.25*0.0 + 0.15*1.0 + 0.15*0.0 + 0.15*1.0 = 0.33
        assert result.score <= 0.35

    def test_assertion_error_with_values_moderate(self) -> None:
        result = run_eval._traceback_informativeness_score(
            exception_type="AssertionError",
            exception_message="assert 1 == 2",
            frame_count=1,
            crash_in_user_code=True,
        )
        # AssertionError has low type specificity (0.1) — even with good
        # assertion quality, the score stays below the 0.7 threshold.
        assert result.score < 0.7
        assert result.f_assertion_quality == 1.0  # comparison values present

    def test_timeout_error_scores_very_low(self) -> None:
        result = run_eval._traceback_informativeness_score(
            exception_type="TimeoutError",
            exception_message="",
            frame_count=5,
            crash_in_user_code=False,
        )
        # 0.30*0.05 + 0.25*0.0 + 0.15*0.0 + 0.15*0.5 + 0.15*0.7 = 0.195
        assert result.score < 0.25

    def test_key_error_scores_moderate(self) -> None:
        result = run_eval._traceback_informativeness_score(
            exception_type="KeyError",
            exception_message="'missing'",
            frame_count=2,
            crash_in_user_code=True,
        )
        assert 0.5 <= result.score <= 0.8

    def test_unknown_exception_default(self) -> None:
        result = run_eval._traceback_informativeness_score(
            exception_type="CustomError",
            exception_message="something happened",
            frame_count=3,
            crash_in_user_code=True,
        )
        assert result.f_type_specificity == run_eval._EXCEPTION_TYPE_SPECIFICITY_DEFAULT

    def test_attribute_error_nonetype(self) -> None:
        result = run_eval._traceback_informativeness_score(
            exception_type="AttributeError",
            exception_message="'NoneType' object has no attribute 'x'",
            frame_count=2,
            crash_in_user_code=True,
        )
        assert 0.55 <= result.score <= 0.85

    def test_deep_stack_reduces_depth_score(self) -> None:
        result = run_eval._traceback_informativeness_score(
            exception_type="TypeError",
            exception_message="unsupported operand type(s)",
            frame_count=8,
            crash_in_user_code=True,
        )
        assert result.f_frame_depth == 0.4  # 7-10 range

    def test_zero_frame_count_gives_neutral_depth(self) -> None:
        result = run_eval._traceback_informativeness_score(
            exception_type="NameError",
            exception_message="name 'x' is not defined",
            frame_count=0,
            crash_in_user_code=False,
        )
        assert result.f_frame_depth == 0.5  # Unknown → neutral

    def test_features_are_populated(self) -> None:
        result = run_eval._traceback_informativeness_score(
            exception_type="SyntaxError",
            exception_message="invalid syntax",
            frame_count=1,
            crash_in_user_code=True,
        )
        assert result.f_type_specificity == 1.0
        assert result.f_message_diagnostic == 1.0
        assert result.f_crash_in_user_code == 1.0
        assert result.f_assertion_quality == 0.5  # neutral for non-assertion
        assert result.f_frame_depth == 1.0


# ===========================================================================
# Unit tests: Extraction
# ===========================================================================


class TestExtractExceptionInfo:
    def test_extract_from_snapshot(self) -> None:
        snapshot = _snapshot_payload("NameError", "name 'x' is not defined")
        exc_type, exc_msg, frame_count, crash_in_user = run_eval._extract_exception_info(
            snapshot=snapshot,
            stdout="",
            stderr="",
        )
        assert exc_type == "NameError"
        assert exc_msg == "name 'x' is not defined"
        assert frame_count == 1
        assert crash_in_user is True

    def test_extract_from_stdout_fallback(self) -> None:
        exc_type, exc_msg, _, _ = run_eval._extract_exception_info(
            snapshot=None,
            stdout="E   NameError: name 'x' is not defined\n",
            stderr="",
        )
        assert exc_type == "NameError"
        assert exc_msg == "name 'x' is not defined"

    def test_extract_from_stderr_fallback(self) -> None:
        exc_type, exc_msg, _, _ = run_eval._extract_exception_info(
            snapshot=None,
            stdout="",
            stderr="NameError: name 'x' is not defined\n",
        )
        assert exc_type == "NameError"
        assert exc_msg == "name 'x' is not defined"

    def test_extract_pytest_format(self) -> None:
        exc_type, exc_msg, _, _ = run_eval._extract_exception_info(
            snapshot=None,
            stdout="E   TypeError: unsupported operand type(s)\n",
            stderr="",
        )
        assert exc_type == "TypeError"
        assert "unsupported" in exc_msg

    def test_extract_returns_unknown_on_empty(self) -> None:
        exc_type, exc_msg, frame_count, crash_in_user = run_eval._extract_exception_info(
            snapshot=None,
            stdout="",
            stderr="",
        )
        assert exc_type == "Unknown"
        assert exc_msg == ""
        assert frame_count == 0
        assert crash_in_user is False

    def test_extract_prefers_snapshot(self) -> None:
        snapshot = _snapshot_payload("SyntaxError", "invalid syntax from snapshot")
        exc_type, exc_msg, _, _ = run_eval._extract_exception_info(
            snapshot=snapshot,
            stdout="E   NameError: name 'y' is not defined\n",
            stderr="",
        )
        assert exc_type == "SyntaxError"
        assert "snapshot" in exc_msg

    def test_extract_strips_module_prefix(self) -> None:
        snapshot = {
            "exception": {"type": "builtins.ValueError", "message": "bad value"},
            "frames": [],
        }
        exc_type, _, _, _ = run_eval._extract_exception_info(
            snapshot=snapshot,
            stdout="",
            stderr="",
        )
        assert exc_type == "ValueError"


# ===========================================================================
# Unit tests: Pre-gate decision
# ===========================================================================


class TestPreGateDecision:
    def test_pregate_skips_syntax_error(self) -> None:
        snapshot = _snapshot_payload("SyntaxError", "invalid syntax")
        result = run_eval._traceback_informativeness_pre_gate(
            snapshot=snapshot,
            stdout="",
            stderr="",
            threshold=0.7,
        )
        assert result.should_skip is True
        assert result.exception_type == "SyntaxError"

    def test_pregate_no_skip_assertion_error(self) -> None:
        snapshot = _snapshot_payload("AssertionError", "")
        result = run_eval._traceback_informativeness_pre_gate(
            snapshot=snapshot,
            stdout="",
            stderr="",
            threshold=0.7,
        )
        assert result.should_skip is False

    def test_pregate_respects_threshold(self) -> None:
        snapshot = _snapshot_payload("NameError", "name 'x' is not defined")
        result = run_eval._traceback_informativeness_pre_gate(
            snapshot=snapshot,
            stdout="",
            stderr="",
            threshold=0.95,
        )
        assert result.should_skip is False
        assert result.score < 0.95

    def test_pregate_returns_features(self) -> None:
        snapshot = _snapshot_payload("IndexError", "list index out of range")
        result = run_eval._traceback_informativeness_pre_gate(
            snapshot=snapshot,
            stdout="",
            stderr="",
            threshold=0.7,
        )
        assert result.f_type_specificity == 0.7
        assert result.f_message_diagnostic == 1.0


# ===========================================================================
# Integration tests
# ===========================================================================


class TestPreGateIntegration:
    def test_adaptive_pregate_skips_stage_a_for_syntax_error(
        self,
        tmp_path: Path,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        case = _make_case(tmp_path)
        workdir = tmp_path / "workdir"
        context_dir = tmp_path / "contexts"
        context_dir.mkdir()
        shutil.copytree(case.path, workdir)
        patcher = _PreGateCapturePatcher()

        monkeypatch.setattr(
            run_eval,
            "run_pytest",
            lambda *args, **kwargs: run_eval.RunResult(
                returncode=1,
                duration_sec=0.01,
                stdout="E   SyntaxError: invalid syntax\n",
                stderr="",
            ),
        )
        monkeypatch.setattr(
            run_eval,
            "read_latest_snapshot",
            lambda *_args, **_kwargs: (
                _snapshot_payload("SyntaxError", "invalid syntax"),
                None,
            ),
        )

        records = run_eval.run_one_condition(
            case=case,
            condition=run_eval.Condition.ADAPTIVE_LLM_DISCRETION,
            workdir=workdir,
            context_dir_base=context_dir,
            patcher=patcher,
            k=1,
            output_format="json_compact",
            run_id="run-pregate-syntax",
            include_files_context=False,
            force_patch_attempt=False,
            adaptive_informativeness_pregate=True,
            adaptive_informativeness_pregate_threshold=0.7,
        )

        assert len(records) >= 1
        assert patcher.request_calls == 0, "Stage-A should NOT have been called"
        meta = records[0]["attempt_metadata"]
        assert meta["adaptive_helper_decision_reason"] == "informativeness_pregate_skip"
        assert meta["adaptive_informativeness_pregate_fired"] is True
        assert meta["adaptive_informativeness_pregate_score"] >= 0.7

    def test_adaptive_pregate_disabled_allows_normal_flow(
        self,
        tmp_path: Path,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        case = _make_case(tmp_path)
        workdir = tmp_path / "workdir"
        context_dir = tmp_path / "contexts"
        context_dir.mkdir()
        shutil.copytree(case.path, workdir)
        patcher = _PreGateCapturePatcher()

        monkeypatch.setattr(
            run_eval,
            "run_pytest",
            lambda *args, **kwargs: run_eval.RunResult(
                returncode=1,
                duration_sec=0.01,
                stdout="E   SyntaxError: invalid syntax\n",
                stderr="",
            ),
        )
        monkeypatch.setattr(
            run_eval,
            "read_latest_snapshot",
            lambda *_args, **_kwargs: (
                _snapshot_payload("SyntaxError", "invalid syntax"),
                None,
            ),
        )

        records = run_eval.run_one_condition(
            case=case,
            condition=run_eval.Condition.ADAPTIVE_LLM_DISCRETION,
            workdir=workdir,
            context_dir_base=context_dir,
            patcher=patcher,
            k=1,
            output_format="json_compact",
            run_id="run-pregate-disabled",
            include_files_context=False,
            force_patch_attempt=False,
            adaptive_informativeness_pregate=False,  # disabled (default)
        )

        assert len(records) >= 1
        # When pregate is disabled, Stage-A should proceed normally
        assert patcher.request_calls >= 1, "Stage-A should have been called with pregate disabled"
        meta = records[0]["attempt_metadata"]
        assert meta["adaptive_informativeness_pregate_fired"] is False

    def test_adaptive_pregate_does_not_fire_for_assertion_error(
        self,
        tmp_path: Path,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        case = _make_case(tmp_path)
        workdir = tmp_path / "workdir"
        context_dir = tmp_path / "contexts"
        context_dir.mkdir()
        shutil.copytree(case.path, workdir)
        patcher = _PreGateCapturePatcher()

        monkeypatch.setattr(
            run_eval,
            "run_pytest",
            lambda *args, **kwargs: run_eval.RunResult(
                returncode=1,
                duration_sec=0.01,
                stdout="E   AssertionError: expected 2 got 1\n",
                stderr="",
            ),
        )
        monkeypatch.setattr(
            run_eval,
            "read_latest_snapshot",
            lambda *_args, **_kwargs: (
                _snapshot_payload("AssertionError", "expected 2 got 1"),
                None,
            ),
        )

        records = run_eval.run_one_condition(
            case=case,
            condition=run_eval.Condition.ADAPTIVE_LLM_DISCRETION,
            workdir=workdir,
            context_dir_base=context_dir,
            patcher=patcher,
            k=1,
            output_format="json_compact",
            run_id="run-pregate-assertion",
            include_files_context=False,
            force_patch_attempt=False,
            adaptive_informativeness_pregate=True,
            adaptive_informativeness_pregate_threshold=0.7,
        )

        assert len(records) >= 1
        # AssertionError scores low, so pregate should NOT fire and Stage-A should proceed
        assert patcher.request_calls >= 1, "Stage-A should be called for AssertionError"
        meta = records[0]["attempt_metadata"]
        assert meta["adaptive_informativeness_pregate_fired"] is False
        assert meta["adaptive_helper_decision_reason"] != "informativeness_pregate_skip"



from __future__ import annotations

from pathlib import Path

from evals.eval_two_pass_summary import main
from llmdebug import _json as json


def _record(
    *,
    run_id: str,
    case_id: str,
    condition: str,
    success: bool,
    schema_version: int = 3,
) -> dict[str, object]:
    return {
        "schema_version": schema_version,
        "run_id": run_id,
        "case_id": case_id,
        "condition": condition,
        "attempt": 1,
        "success": success,
        "category": "logic_error",
        "difficulty": "easy",
        "snapshot_dependency": "required",
        "run_metadata": {
            "run_id": run_id,
            "started_at_utc": "2026-03-01T00:00:00Z",
            "deterministic": False,
            "seed": None,
            "python_version": "3.11.0",
            "platform": "test",
            "llmdebug_version": "2.11.4",
            "patcher_name": "openai",
            "patcher_config_fingerprint": "fp",
        },
        "attempt_metadata": {
            "attempt": 1,
            "max_attempts": 3,
            "attempt_ordinal_label": "1/3",
            "llm_call_count": 1,
            "tokens_prompt": 10,
            "tokens_completion": 2,
            "tokens_total": 12,
            "retry_history_budget_chars": 12000,
            "retry_history_budget_applied": False,
            "prompt_sha256": "prompt-hash",
            "prompt_path": None,
            "evidence_sha256": "evidence-hash",
            "evidence_path": None,
        },
        "model_metadata": {
            "backend": "openai",
            "name": "openai",
            "base_url": "http://127.0.0.1:12000/v1",
            "model": "dummy",
            "temperature": 0.0,
            "max_tokens": 2048,
            "response_mode": "edits",
            "api_compat_seed_requested": None,
            "api_compat_seed_acknowledged": False,
        },
        "cumulative_runtime_sec": 1.0,
        "note": "success" if success else "tests_still_failing",
    }


def _write_jsonl(path: Path, records: list[dict[str, object]]) -> None:
    path.write_text(
        "\n".join(json.dumps(record, ensure_ascii=True) for record in records) + "\n",
        encoding="utf-8",
    )


def test_two_pass_summary_missing_file(tmp_path: Path, capsys) -> None:
    rc = main(["base", "rescue", "--results-dir", str(tmp_path)])
    assert rc == 1
    assert "results file not found" in capsys.readouterr().out


def test_two_pass_summary_strict_schema_rejection(tmp_path: Path, capsys) -> None:
    _write_jsonl(
        tmp_path / "base.jsonl",
        [_record(run_id="base", case_id="case-a", condition="traceback_only", success=False)],
    )
    _write_jsonl(
        tmp_path / "rescue.jsonl",
        [
            _record(
                run_id="rescue",
                case_id="case-a",
                condition="adaptive_llm_discretion",
                success=True,
                schema_version=2,
            )
        ],
    )
    rc = main(["base", "rescue", "--results-dir", str(tmp_path)])
    assert rc == 2
    assert "unsupported schema_version=2" in capsys.readouterr().out


def test_two_pass_summary_computes_rescue_uplift(tmp_path: Path, capsys) -> None:
    _write_jsonl(
        tmp_path / "base.jsonl",
        [
            _record(run_id="base", case_id="case-a", condition="traceback_only", success=False),
            _record(run_id="base", case_id="case-b", condition="traceback_only", success=False),
            _record(run_id="base", case_id="case-c", condition="traceback_only", success=True),
        ],
    )
    _write_jsonl(
        tmp_path / "rescue.jsonl",
        [
            _record(
                run_id="rescue",
                case_id="case-a",
                condition="adaptive_llm_discretion",
                success=True,
            ),
            _record(
                run_id="rescue",
                case_id="case-b",
                condition="adaptive_llm_discretion",
                success=False,
            ),
        ],
    )
    rc = main(["base", "rescue", "--results-dir", str(tmp_path)])
    assert rc == 0
    out = capsys.readouterr().out
    assert "- success: 1/3 (33.3%)" in out
    assert "- attempted failed cases: 2/2 (100.0%)" in out
    assert "1/2 on attempted (50.0%), 1/2 on failed pool (50.0%)" in out
    assert "- final success: 2/3 (66.7%)" in out
    assert "- absolute uplift vs baseline: +33.3%" in out


def test_two_pass_summary_handles_no_failed_baseline(tmp_path: Path, capsys) -> None:
    _write_jsonl(
        tmp_path / "base.jsonl",
        [_record(run_id="base", case_id="case-a", condition="traceback_only", success=True)],
    )
    _write_jsonl(
        tmp_path / "rescue.jsonl",
        [
            _record(
                run_id="rescue",
                case_id="case-a",
                condition="adaptive_llm_discretion",
                success=True,
            )
        ],
    )
    rc = main(["base", "rescue", "--results-dir", str(tmp_path)])
    assert rc == 0
    out = capsys.readouterr().out
    assert "- failed cases: 0" in out
    assert "- attempted failed cases: 0/0 (n/a)" in out
    assert "- final success: 1/1 (100.0%)" in out
    assert "- absolute uplift vs baseline: +0.0%" in out


def test_two_pass_summary_rejects_mixed_run_ids_in_input_file(tmp_path: Path, capsys) -> None:
    _write_jsonl(
        tmp_path / "base.jsonl",
        [
            _record(run_id="base", case_id="case-a", condition="traceback_only", success=False),
            _record(run_id="other", case_id="case-b", condition="traceback_only", success=True),
        ],
    )
    _write_jsonl(
        tmp_path / "rescue.jsonl",
        [
            _record(
                run_id="rescue",
                case_id="case-a",
                condition="adaptive_llm_discretion",
                success=True,
            )
        ],
    )
    rc = main(["base", "rescue", "--results-dir", str(tmp_path)])
    assert rc == 2
    out = capsys.readouterr().out
    assert "must contain only run_id='base'" in out



"""Identity/signature helpers for RCA workflows."""

from __future__ import annotations

import hashlib
from pathlib import Path
from typing import Any, cast

from . import _json as json
from ._debug_session import coerce_crash_frame_index, normalize_snapshot_view


def _signature_from_snapshot(view: dict[str, Any]) -> str:
    """Build failure signature from a normalized snapshot view."""
    exc = view.get("exception", {})
    exc_type = str(exc.get("type") or "Exception")
    exc_message = str(exc.get("message") or "")
    frames_raw = view.get("frames")
    frames: list[dict[str, Any]] = (
        cast(list[dict[str, Any]], frames_raw) if isinstance(frames_raw, list) else []
    )
    crash_frame: dict[str, Any] = {}
    if frames:
        crash_idx = coerce_crash_frame_index(view)
        candidate = frames[crash_idx] if 0 <= crash_idx < len(frames) else frames[0]
        if isinstance(candidate, dict):  # pyright: ignore[reportUnnecessaryIsInstance] — defense-in-depth for malformed data
            crash_frame = candidate
    location = {
        "file": str(crash_frame.get("file_rel") or crash_frame.get("file") or "?"),
        "line": int(crash_frame.get("line") or 0),
        "function": str(crash_frame.get("function") or "?"),
    }
    nodeid = ""
    pytest_block = view.get("pytest")
    if isinstance(pytest_block, dict):
        nodeid = str(cast(dict[str, Any], pytest_block).get("nodeid") or "")
    payload = {
        "exception_type": exc_type,
        "exception_message": exc_message,
        "location": location,
        "nodeid": nodeid,
    }
    digest = hashlib.sha256(
        json.dumps(payload, ensure_ascii=False, sort_keys=True, separators=(",", ":")).encode(
            "utf-8"
        )
    ).hexdigest()[:16]
    return f"{exc_type}:{digest}"


def _signature_from_text(stdout: str, stderr: str) -> str:
    """Build failure signature from raw text output."""
    text = (stderr or "").strip() or (stdout or "").strip() or "unknown_failure"
    lines = [ln.strip() for ln in text.splitlines() if ln.strip()]
    head = "\n".join(lines[-3:]) if lines else text
    digest = hashlib.sha256(head.encode("utf-8")).hexdigest()[:16]
    return f"TextFailure:{digest}"


def build_failure_signature(
    snapshot: dict[str, Any] | None,
    *,
    stdout: str = "",
    stderr: str = "",
) -> str:
    """Build a stable failure signature from snapshot (preferred) or textual output."""
    if snapshot:
        return _signature_from_snapshot(normalize_snapshot_view(snapshot))
    return _signature_from_text(stdout, stderr)


def derive_session_id(
    snapshot: dict[str, Any] | None,
    *,
    out_dir: str,
    fallback: str = "global",
) -> str:
    """Derive stable session id from out_dir + pytest nodeid/name when available."""
    resolved = str(Path(out_dir).resolve())
    scope = fallback
    if snapshot:
        view = normalize_snapshot_view(snapshot)
        pytest_block = view.get("pytest")
        if isinstance(pytest_block, dict):
            pytest_dict = cast(dict[str, Any], pytest_block)
            raw_nodeid = pytest_dict.get("nodeid")
            if isinstance(raw_nodeid, str):
                nodeid = raw_nodeid.strip()
                if nodeid:
                    scope = nodeid
        if scope == fallback and isinstance(view.get("name"), str):
            name = str(view.get("name")).strip()
            if name:
                scope = name
    return f"{resolved}::{scope}"



"""Patcher interface for eval harness.

A patcher receives a prompt (plus a context directory path) and returns a single
unified diff that can be applied with `git apply`.
"""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

from .payload_types import ContextRequestPayload, PatcherMeta


@dataclass(frozen=True)
class PatchResult:
    diff: str | None
    meta: PatcherMeta | None = None


@dataclass(frozen=True)
class ContextRequestResult:
    request_json: ContextRequestPayload | None
    raw_response: str | None = None
    meta: PatcherMeta | None = None


@dataclass(frozen=True)
class ContextCompactionResult:
    compacted_text: str | None
    raw_response: str | None = None
    meta: PatcherMeta | None = None


class Patcher:
    name: str

    def request_context(
        self,
        *,
        prompt: str,
        context_dir: Path,
        cwd: Path,
    ) -> ContextRequestResult | None:
        return None

    async def arequest_context(
        self,
        *,
        prompt: str,
        context_dir: Path,
        cwd: Path,
    ) -> ContextRequestResult | None:
        return self.request_context(prompt=prompt, context_dir=context_dir, cwd=cwd)

    def compact_context(
        self,
        *,
        prompt: str,
        context_text: str,
        context_dir: Path,
        cwd: Path,
        target_chars: int,
    ) -> ContextCompactionResult | None:
        return None

    async def acompact_context(
        self,
        *,
        prompt: str,
        context_text: str,
        context_dir: Path,
        cwd: Path,
        target_chars: int,
    ) -> ContextCompactionResult | None:
        return self.compact_context(
            prompt=prompt,
            context_text=context_text,
            context_dir=context_dir,
            cwd=cwd,
            target_chars=target_chars,
        )

    def propose_patch(self, *, prompt: str, context_dir: Path, cwd: Path) -> PatchResult:
        raise NotImplementedError

    async def apropose_patch(self, *, prompt: str, context_dir: Path, cwd: Path) -> PatchResult:
        return self.propose_patch(prompt=prompt, context_dir=context_dir, cwd=cwd)



import subprocess
import sys


def test_condefects_0():
    proc = subprocess.run(
        [sys.executable, "buggy.py"],
        input='3\nabc\nabb\naac\n',
        capture_output=True,
        text=True,
        timeout=10,
    )
    expected = '2\n2\n1\n'
    actual = proc.stdout.strip()
    assert actual == expected.strip(), (
        f"Expected {expected.strip()!r}, got {actual!r}"
    )


def test_condefects_1():
    proc = subprocess.run(
        [sys.executable, "buggy.py"],
        input='11\nabracadabra\nbracadabra\nracadabra\nacadabra\ncadabra\nadabra\ndabra\nabra\nbra\nra\na\n',
        capture_output=True,
        text=True,
        timeout=10,
    )
    expected = '4\n3\n2\n1\n0\n1\n0\n4\n3\n2\n1\n'
    actual = proc.stdout.strip()
    assert actual == expected.strip(), (
        f"Expected {expected.strip()!r}, got {actual!r}"
    )



"""Persistence adapters for RCA state (sqlite/jsonl)."""

from __future__ import annotations

import os
from collections.abc import Iterator
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Literal, cast

from filelock import FileLock

from . import _json as json
from .rca_store import (
    has_any_records,
    history,
    latest_record,
    store_rca_record,
    store_rca_records_bulk,
)

RCABackend = Literal["sqlite", "jsonl"]

_RCA_FILE = "rca_attempts.jsonl"
_RCA_LOCK_FILE = ".llmdebug.rca.lock"
_RCA_BACKEND_ENV = "LLMDEBUG_RCA_BACKEND"
_DEFAULT_BACKEND: RCABackend = "sqlite"
_DEFAULT_MAX_RECORDS = 5000
_REQUIRED_RCA_FIELDS = (
    "attempt_id",
    "session_id",
    "updated_at",
    "created_at",
    "failure_signature",
    "state",
)


def _store_path(out_dir: str) -> Path:
    return Path(out_dir) / _RCA_FILE


@contextmanager
def _store_lock(path: Path) -> Iterator[None]:
    """Serialize RCA writes and migration operations."""
    lock = FileLock(path.parent / _RCA_LOCK_FILE)
    with lock:
        yield


def _backend() -> RCABackend:
    raw = os.getenv(_RCA_BACKEND_ENV, _DEFAULT_BACKEND).strip().lower()
    if raw in {"sqlite", "jsonl"}:
        return cast(RCABackend, raw)
    return _DEFAULT_BACKEND


def max_rca_records() -> int:
    """Return configured max retained RCA records (0 means unlimited)."""
    raw_value = os.getenv("LLMDEBUG_RCA_MAX_RECORDS", str(_DEFAULT_MAX_RECORDS))
    try:
        value = int(raw_value)
    except ValueError:
        return _DEFAULT_MAX_RECORDS
    return max(value, 0)


def _iter_jsonl_records(path: Path) -> list[dict[str, Any]]:
    if not path.exists():
        return []
    records: list[dict[str, Any]] = []
    try:
        raw = path.read_text(encoding="utf-8")
    except OSError:
        return []
    for line in raw.splitlines():
        item = line.strip()
        if not item:
            continue
        try:
            data = json.loads(item)
        except json.JSONDecodeError:
            continue
        if isinstance(data, dict):
            records.append(cast(dict[str, Any], data))
    return records


def _is_valid_rca_record(record: dict[str, Any]) -> bool:
    for key in _REQUIRED_RCA_FIELDS:
        value = record.get(key)
        if value is None or not str(value).strip():
            return False
    return True


def _iter_jsonl_lines_reverse(path: Path, *, block_size: int = 8192) -> Iterator[str]:
    """Yield JSONL lines in reverse order without loading the whole file."""
    with path.open("rb") as handle:
        handle.seek(0, os.SEEK_END)
        position = handle.tell()
        pending = b""

        while position > 0:
            read_size = min(block_size, position)
            position -= read_size
            handle.seek(position)
            chunk = handle.read(read_size)
            pending = chunk + pending
            lines = pending.split(b"\n")
            pending = lines[0]
            for raw_line in reversed(lines[1:]):
                line = raw_line.strip()
                if line:
                    yield line.decode("utf-8", errors="replace")

        if pending.strip():
            yield pending.strip().decode("utf-8", errors="replace")


def _latest_jsonl_record(path: Path, *, session_id: str | None = None) -> dict[str, Any] | None:
    if not path.exists():
        return None
    try:
        for line in _iter_jsonl_lines_reverse(path):
            try:
                data = json.loads(line)
            except json.JSONDecodeError:
                continue
            if not isinstance(data, dict):
                continue
            data_dict = cast(dict[str, Any], data)
            if session_id and data_dict.get("session_id") != session_id:
                continue
            return data_dict
    except OSError:
        return None
    return None


def _append_record(path: Path, record: dict[str, Any]) -> None:
    path.parent.mkdir(parents=True, exist_ok=True)
    with _store_lock(path):
        with path.open("a", encoding="utf-8") as handle:
            handle.write(json.dumps(record, ensure_ascii=False, sort_keys=True) + "\n")

        max_records = max_rca_records()
        if max_records <= 0:
            return

        try:
            lines = path.read_text(encoding="utf-8").splitlines()
        except OSError:
            return
        if len(lines) <= max_records:
            return

        trimmed = lines[-max_records:]
        try:
            path.write_text("\n".join(trimmed) + "\n", encoding="utf-8")
        except OSError:
            return


def ensure_sqlite_migration(out_dir: str) -> None:
    """Best-effort one-way migration from JSONL store to sqlite store."""
    jsonl_path = _store_path(out_dir)
    if not jsonl_path.exists():
        return

    with _store_lock(jsonl_path):
        if not jsonl_path.exists():
            return
        if has_any_records(out_dir=out_dir):
            return

        records = _iter_jsonl_records(jsonl_path)
        if records:
            valid_records = [record for record in records if _is_valid_rca_record(record)]
            if valid_records:
                store_rca_records_bulk(
                    valid_records,
                    out_dir=out_dir,
                    max_records=max_rca_records(),
                )

        backup_path = Path(f"{jsonl_path}.bak")
        try:
            backup_path.unlink(missing_ok=True)
        except OSError:
            pass
        try:
            jsonl_path.replace(backup_path)
        except OSError:
            # Keep original JSONL if backup rotation fails.
            pass


def get_rca_history(
    *,
    out_dir: str,
    session_id: str | None = None,
    limit: int = 20,
) -> list[dict[str, Any]]:
    """Load RCA history, newest-first."""
    if _backend() == "sqlite":
        ensure_sqlite_migration(out_dir)
        return history(out_dir=out_dir, session_id=session_id, limit=limit)

    records = _iter_jsonl_records(_store_path(out_dir))
    if session_id:
        records = [record for record in records if record.get("session_id") == session_id]
    records = list(reversed(records))
    if limit > 0:
        records = records[:limit]
    return records


def get_rca_status(*, out_dir: str, session_id: str | None = None) -> dict[str, Any] | None:
    """Return latest RCA attempt for a session."""
    if _backend() == "sqlite":
        ensure_sqlite_migration(out_dir)
        return latest_record(out_dir=out_dir, session_id=session_id)

    return _latest_jsonl_record(_store_path(out_dir), session_id=session_id)


def persist_rca_record(*, out_dir: str, record: dict[str, Any]) -> None:
    """Persist one RCA record using the configured backend."""
    if _backend() == "sqlite":
        ensure_sqlite_migration(out_dir)
        store_rca_record(record, out_dir=out_dir, max_records=max_rca_records())
        return

    _append_record(_store_path(out_dir), record)


def backend_name() -> RCABackend:
    """Expose backend name for targeted tests/diagnostics."""
    return _backend()



"""Git repository context capture for debug snapshots."""

from __future__ import annotations

import re
import subprocess
import time
from copy import deepcopy
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from threading import Lock
from typing import Any, cast

from .config import GitDirtyMode


class _GitError:
    """Sentinel to distinguish 'command failed' from 'not applicable'."""

    __slots__ = ("reason",)

    def __init__(self, reason: str) -> None:
        self.reason = reason


@dataclass(frozen=True)
class _GitContextCacheEntry:
    """Small in-process cache for git context lookups."""

    expires_at: float
    cwd: str | None
    repo_root: str | None
    git_dir: str | None
    head_mtime_ns: int | None
    index_mtime_ns: int | None
    dirty_cached: bool | None
    context_static: dict[str, Any] | None


_GIT_CONTEXT_CACHE_TTL_SEC = 1.0
_GIT_CONTEXT_CACHE_LOCK = Lock()
_git_context_cache: _GitContextCacheEntry | None = None


def _run_git_command(
    args: list[str],
    timeout: float = 2.0,
    repo_root: str | None = None,
) -> str | _GitError:
    """Run a git command with timeout.

    Returns:
        stdout string on success, _GitError on failure.
    """
    cmd = ["git"]
    if repo_root:
        cmd.extend(["-C", repo_root])
    cmd.extend(args)
    try:
        result = subprocess.run(
            cmd,
            capture_output=True,
            text=True,
            timeout=timeout,
            check=False,
        )
        if result.returncode == 0:
            return result.stdout.strip()
        return _GitError(f"exit code {result.returncode}")
    except subprocess.TimeoutExpired:
        return _GitError("timeout")
    except FileNotFoundError:
        return _GitError("git not found")
    except OSError as e:
        return _GitError(f"OSError: {e}")


def _run_git_command_rc(
    args: list[str],
    timeout: float = 2.0,
    repo_root: str | None = None,
) -> int | _GitError:
    """Run a git command and return process exit code."""
    cmd = ["git"]
    if repo_root:
        cmd.extend(["-C", repo_root])
    cmd.extend(args)
    try:
        result = subprocess.run(
            cmd,
            capture_output=True,
            text=True,
            timeout=timeout,
            check=False,
        )
        return result.returncode
    except subprocess.TimeoutExpired:
        return _GitError("timeout")
    except FileNotFoundError:
        return _GitError("git not found")
    except OSError as e:
        return _GitError(f"OSError: {e}")


def _resolve_cwd() -> str | None:
    try:
        return str(Path.cwd().resolve())
    except (OSError, RuntimeError, ValueError):
        return None


def _path_within_repo(path: str, repo_root: str) -> bool:
    try:
        Path(path).resolve().relative_to(Path(repo_root).resolve())
    except (OSError, RuntimeError, ValueError):
        return False
    return True


def _resolve_git_dir(repo_root: str) -> Path | None:
    git_entry = Path(repo_root) / ".git"
    try:
        if git_entry.is_dir():
            return git_entry.resolve()
        if not git_entry.is_file():
            return None

        raw = git_entry.read_text(encoding="utf-8", errors="replace").strip()
        if not raw:
            return None
        prefix = "gitdir:"
        if not raw.lower().startswith(prefix):
            return None
        target_raw = raw[len(prefix) :].strip()
        if not target_raw:
            return None

        git_dir = Path(target_raw)
        if not git_dir.is_absolute():
            git_dir = (git_entry.parent / git_dir).resolve()
        else:
            git_dir = git_dir.resolve()
        return git_dir
    except (OSError, RuntimeError, ValueError):
        return None


def _git_signal_mtimes(repo_root: str) -> tuple[str | None, int | None, int | None]:
    git_dir = _resolve_git_dir(repo_root)
    if git_dir is None:
        return None, None, None

    def _mtime_ns(path: Path) -> int | None:
        try:
            return path.stat().st_mtime_ns
        except OSError:
            return None

    return (
        str(git_dir),
        _mtime_ns(git_dir / "HEAD"),
        _mtime_ns(git_dir / "index"),
    )


def _detect_untracked_status(repo_root: str) -> bool | None:
    """Return untracked-file status using `ls-files` exit codes."""
    untracked = _run_git_command_rc(
        [
            "ls-files",
            "--others",
            "--exclude-standard",
            "--directory",
            "--no-empty-directory",
            "--error-unmatch",
            "--",
            "*",
        ],
        repo_root=repo_root,
    )
    if isinstance(untracked, _GitError):
        return None
    if untracked == 0:
        return True
    if untracked == 1:
        return False
    return None


def _detect_dirty_status(repo_root: str) -> bool | None:
    """Return dirty status with a faster command pair than `status --porcelain`."""
    unstaged = _run_git_command_rc(
        ["diff-files", "--quiet", "--ignore-submodules", "--"],
        repo_root=repo_root,
    )
    if isinstance(unstaged, _GitError):
        return None
    if unstaged == 1:
        return True
    if unstaged != 0:
        return None

    staged = _run_git_command_rc(
        ["diff-index", "--quiet", "--cached", "HEAD", "--"],
        repo_root=repo_root,
    )
    if isinstance(staged, _GitError):
        return None
    if staged == 1:
        return True
    if staged == 0:
        return _detect_untracked_status(repo_root)
    return None


def _hydrate_context_with_dirty(context_static: dict[str, Any], *, repo_root: str) -> dict[str, Any]:
    context = deepcopy(context_static)

    dirty = _detect_dirty_status(repo_root)
    if dirty is None:
        status = _run_git_command(["status", "--porcelain"], repo_root=repo_root)
        if isinstance(status, str):
            dirty = len(status) > 0
    if dirty is not None:
        context["dirty"] = dirty
    return context


def _load_cached_git_context(*, dirty_mode: GitDirtyMode) -> tuple[bool, dict[str, Any] | None]:
    now = time.monotonic()
    with _GIT_CONTEXT_CACHE_LOCK:
        entry = _git_context_cache
    if entry is None or now >= entry.expires_at:
        return False, None

    cwd = _resolve_cwd()
    if entry.repo_root is None:
        if cwd != entry.cwd:
            return False, None
        return True, None

    if cwd is None or not _path_within_repo(cwd, entry.repo_root):
        return False, None

    git_dir, head_mtime_ns, index_mtime_ns = _git_signal_mtimes(entry.repo_root)
    if (
        git_dir != entry.git_dir
        or head_mtime_ns != entry.head_mtime_ns
        or index_mtime_ns != entry.index_mtime_ns
    ):
        return False, None

    if entry.context_static is None:
        return True, None

    if dirty_mode == "off":
        return True, deepcopy(entry.context_static)

    if dirty_mode == "cached":
        if entry.dirty_cached is None:
            return False, None
        context = deepcopy(entry.context_static)
        context["dirty"] = entry.dirty_cached
        return True, context

    return True, _hydrate_context_with_dirty(entry.context_static, repo_root=entry.repo_root)


def _store_git_context_cache(
    context: dict[str, Any] | None,
    *,
    repo_root: str | None,
) -> None:
    cwd = _resolve_cwd()
    git_dir, head_mtime_ns, index_mtime_ns = (None, None, None)
    dirty_cached: bool | None = None
    context_static: dict[str, Any] | None = None

    if repo_root is not None:
        git_dir, head_mtime_ns, index_mtime_ns = _git_signal_mtimes(repo_root)
        if context:
            context_static = deepcopy(context)
            dirty_value = context_static.pop("dirty", None)
            if isinstance(dirty_value, bool):
                dirty_cached = dirty_value

    entry = _GitContextCacheEntry(
        expires_at=time.monotonic() + max(0.0, _GIT_CONTEXT_CACHE_TTL_SEC),
        cwd=cwd,
        repo_root=repo_root,
        git_dir=git_dir,
        head_mtime_ns=head_mtime_ns,
        index_mtime_ns=index_mtime_ns,
        dirty_cached=dirty_cached,
        context_static=context_static,
    )
    with _GIT_CONTEXT_CACHE_LOCK:
        global _git_context_cache
        _git_context_cache = entry


def _remaining_budget(deadline: float) -> float:
    return max(0.0, deadline - time.monotonic())


def _run_git_budgeted(
    args: list[str],
    *,
    repo_root: str | None,
    label: str,
    warnings: list[str],
    per_cmd_timeout_sec: float,
    deadline: float,
) -> str | None:
    remaining = _remaining_budget(deadline)
    if remaining <= 0:
        warnings.append(f"{label}: budget exhausted")
        return None

    timeout = min(max(per_cmd_timeout_sec, 0.01), remaining)
    result = _run_git_command(args, timeout=timeout, repo_root=repo_root)
    if isinstance(result, str):
        return result
    warnings.append(f"{label}: {result.reason}")
    return None


def _extract_crash_frame(snapshot: dict[str, Any]) -> dict[str, Any] | None:
    frames = snapshot.get("frames")
    if not isinstance(frames, list) or not frames:
        return None
    frames_list = cast(list[dict[str, Any]], frames)

    crash_idx_raw = snapshot.get("crash_frame_index", 0)
    try:
        crash_idx = int(crash_idx_raw)
    except (TypeError, ValueError):
        crash_idx = 0

    if 0 <= crash_idx < len(frames_list) and isinstance(frames_list[crash_idx], dict):  # pyright: ignore[reportUnnecessaryIsInstance]
        return frames_list[crash_idx]
    if isinstance(frames_list[0], dict):  # pyright: ignore[reportUnnecessaryIsInstance] — defense-in-depth for malformed data
        return frames_list[0]
    return None


def _extract_crash_line(crash_frame: dict[str, Any]) -> int | None:
    line_raw = crash_frame.get("line")
    if isinstance(line_raw, int):
        line = line_raw
    elif isinstance(line_raw, str):
        try:
            line = int(line_raw)
        except ValueError:
            return None
    else:
        return None
    return line if line > 0 else None


def _resolve_repo_relative_file(
    crash_frame: dict[str, Any],
    *,
    repo_root: str,
) -> tuple[str | None, bool]:
    repo_root_path = Path(repo_root).resolve()

    file_rel_raw = crash_frame.get("file_rel")
    if isinstance(file_rel_raw, str) and file_rel_raw:
        rel_candidate = file_rel_raw.replace("\\", "/")
        if not rel_candidate.startswith(".."):
            try:
                joined = (repo_root_path / rel_candidate).resolve()
                rel = joined.relative_to(repo_root_path).as_posix()
                return rel, True
            except (OSError, ValueError):
                pass
        return rel_candidate, False

    file_abs_raw = crash_frame.get("file")
    if isinstance(file_abs_raw, str) and file_abs_raw:
        try:
            file_abs = Path(file_abs_raw).resolve()
            rel = file_abs.relative_to(repo_root_path).as_posix()
            return rel, True
        except (OSError, ValueError):
            return file_abs_raw, False

    return None, False


def _parse_blame_porcelain(text: str, line: int) -> dict[str, Any] | None:
    if not text:
        return None

    result: dict[str, Any] = {"line": line}
    for raw_line in text.splitlines():
        line_text = raw_line.strip()
        if not line_text:
            continue

        if "commit" not in result and re.match(r"^[0-9a-f]{40,64}\s+\d+\s+\d+\s+\d+$", line_text):
            result["commit"] = line_text.split()[0]
            continue

        if line_text.startswith("author "):
            result["author"] = line_text.removeprefix("author ").strip()
            continue

        if line_text.startswith("author-time "):
            ts_raw = line_text.removeprefix("author-time ").strip()
            try:
                ts = int(ts_raw)
                result["author_time"] = (
                    datetime.fromtimestamp(ts, tz=timezone.utc).isoformat().replace("+00:00", "Z")
                )
            except ValueError:
                result["author_time"] = ts_raw
            continue

        if line_text.startswith("summary "):
            result["summary"] = line_text.removeprefix("summary ").strip()
            continue

    if "commit" not in result:
        return None
    return result


_SHORTSTAT_RE = re.compile(
    r"(?P<files>\d+)\s+files?\s+changed"
    r"(?:,\s+(?P<insertions>\d+)\s+insertions?\(\+\))?"
    r"(?:,\s+(?P<deletions>\d+)\s+deletions?\(-\))?"
)
_LOG_SEPARATOR = "__LLMDEBUG_COMMIT__"


def _parse_shortstat(line: str) -> tuple[int, int, int]:
    m = _SHORTSTAT_RE.search(line)
    if m is None:
        return 0, 0, 0

    files_changed = int(m.group("files") or 0)
    insertions = int(m.group("insertions") or 0)
    deletions = int(m.group("deletions") or 0)
    return files_changed, insertions, deletions


def _parse_recent_commits(log_text: str) -> list[dict[str, Any]]:
    if not log_text:
        return []

    lines = [line.rstrip() for line in log_text.splitlines()]
    commits: list[dict[str, Any]] = []
    i = 0
    while i < len(lines):
        if lines[i].strip() != _LOG_SEPARATOR:
            i += 1
            continue

        if i + 5 >= len(lines):
            break

        short_commit = lines[i + 1].strip()
        full_commit = lines[i + 2].strip()
        author = lines[i + 3].strip()
        author_time = lines[i + 4].strip()
        summary = lines[i + 5].strip()

        i += 6
        files_changed = 0
        insertions = 0
        deletions = 0
        while i < len(lines) and lines[i].strip() != _LOG_SEPARATOR:
            line = lines[i].strip()
            if "file changed" in line or "files changed" in line:
                files_changed, insertions, deletions = _parse_shortstat(line)
                break
            i += 1

        commits.append(
            {
                "commit": short_commit,
                "commit_full": full_commit,
                "author": author,
                "author_time": author_time,
                "summary": summary,
                "files_changed": files_changed,
                "insertions": insertions,
                "deletions": deletions,
            }
        )

    return commits


def _parse_numstat(text: str) -> dict[str, Any]:
    files_changed = 0
    insertions = 0
    deletions = 0

    for raw_line in text.splitlines():
        line = raw_line.strip()
        if not line:
            continue
        parts = line.split("\t")
        if len(parts) < 3:
            continue
        files_changed += 1
        ins_raw, del_raw = parts[0], parts[1]
        if ins_raw.isdigit():
            insertions += int(ins_raw)
        if del_raw.isdigit():
            deletions += int(del_raw)

    return {
        "files_changed": files_changed,
        "insertions": insertions,
        "deletions": deletions,
        "touched": files_changed > 0,
    }


def _candidate_repo_dirs(snapshot: dict[str, Any], *, out_dir: str) -> list[str]:
    """Build an ordered candidate list to discover the snapshot's git repo."""
    candidates: list[str] = []
    seen: set[str] = set()

    def _add(raw_path: str | Path | None, *, from_file: bool = False) -> None:
        if raw_path is None:
            return
        raw_text = str(raw_path).strip()
        if not raw_text:
            return

        try:
            path = Path(raw_text).expanduser()
            if from_file:
                path = path.parent
            resolved = path.resolve()
        except (OSError, RuntimeError, ValueError):
            return

        key = resolved.as_posix()
        if key in seen:
            return
        seen.add(key)
        candidates.append(key)

    crash_frame = _extract_crash_frame(snapshot)
    if crash_frame is not None:
        file_abs_raw = crash_frame.get("file")
        if isinstance(file_abs_raw, str) and file_abs_raw:
            file_path = Path(file_abs_raw)
            if file_path.is_absolute():
                _add(file_path, from_file=True)

    _add(out_dir)

    try:
        _add(Path(out_dir).expanduser().resolve().parent)
    except (OSError, RuntimeError, ValueError):
        pass

    env = snapshot.get("env")
    if isinstance(env, dict):
        env_dict = cast(dict[str, Any], env)
        cwd_raw = env_dict.get("cwd")
        if isinstance(cwd_raw, str) and cwd_raw:
            _add(cwd_raw)
    return candidates


def _discover_repo_root(
    snapshot: dict[str, Any],
    *,
    out_dir: str,
    warnings: list[str],
    per_cmd_timeout_sec: float,
    deadline: float,
) -> str | None:
    candidates = _candidate_repo_dirs(snapshot, out_dir=out_dir)
    if not candidates:
        warnings.append("repo root: no discovery candidates")
        return None

    failures: list[str] = []
    for candidate in candidates:
        remaining = _remaining_budget(deadline)
        if remaining <= 0:
            warnings.append("repo root: budget exhausted")
            return None

        timeout = min(max(per_cmd_timeout_sec, 0.01), remaining)
        result = _run_git_command(
            ["rev-parse", "--show-toplevel"],
            timeout=timeout,
            repo_root=candidate,
        )
        if isinstance(result, str):
            return result
        failures.append(f"{candidate} ({result.reason})")

    preview = "; ".join(failures[:2])
    if len(failures) > 2:
        preview = f"{preview}; +{len(failures) - 2} more"
    warnings.append(f"repo root: not found from candidates: {preview}")
    return None


def get_enhanced_git_context(
    snapshot: dict[str, Any],
    *,
    out_dir: str,
    max_commits: int = 5,
    per_cmd_timeout_sec: float = 0.4,
    total_budget_sec: float = 1.2,
) -> dict[str, Any] | None:
    """Compute enhanced git metadata on-demand for a specific snapshot.

    This is intentionally metadata-only and budgeted:
    - crash-line blame metadata (no code hunk)
    - recent commit metadata + shortstats
    - crash-file diffstat metadata
    """
    warnings: list[str] = []
    deadline = time.monotonic() + max(total_budget_sec, 0.0)

    repo_root = _discover_repo_root(
        snapshot,
        out_dir=out_dir,
        warnings=warnings,
        per_cmd_timeout_sec=per_cmd_timeout_sec,
        deadline=deadline,
    )
    if repo_root is None:
        return None

    context: dict[str, Any] = {"repo_root": repo_root}
    head: dict[str, Any] = {}

    commit_short = _run_git_budgeted(
        ["rev-parse", "--short", "HEAD"],
        repo_root=repo_root,
        label="head commit",
        warnings=warnings,
        per_cmd_timeout_sec=per_cmd_timeout_sec,
        deadline=deadline,
    )
    if commit_short is not None:
        head["commit"] = commit_short

    commit_full = _run_git_budgeted(
        ["rev-parse", "HEAD"],
        repo_root=repo_root,
        label="head commit full",
        warnings=warnings,
        per_cmd_timeout_sec=per_cmd_timeout_sec,
        deadline=deadline,
    )
    if commit_full is not None:
        head["commit_full"] = commit_full

    branch = _run_git_budgeted(
        ["rev-parse", "--abbrev-ref", "HEAD"],
        repo_root=repo_root,
        label="branch",
        warnings=warnings,
        per_cmd_timeout_sec=per_cmd_timeout_sec,
        deadline=deadline,
    )
    if branch is not None:
        head["branch"] = branch
        if branch == "HEAD":
            head["detached"] = True

    status = _run_git_budgeted(
        ["status", "--porcelain"],
        repo_root=repo_root,
        label="dirty status",
        warnings=warnings,
        per_cmd_timeout_sec=per_cmd_timeout_sec,
        deadline=deadline,
    )
    if status is not None:
        head["dirty"] = len(status) > 0

    if head:
        context["head"] = head

    crash_frame = _extract_crash_frame(snapshot)
    if crash_frame is not None:
        crash_line = _extract_crash_line(crash_frame)
        file_rel, in_repo = _resolve_repo_relative_file(crash_frame, repo_root=repo_root)
        crash_file: dict[str, Any] = {"file_rel": file_rel, "in_repo": in_repo}
        if crash_line is not None:
            crash_file["line"] = crash_line

        if file_rel and in_repo and crash_line is not None:
            blame_text = _run_git_budgeted(
                ["blame", "-L", f"{crash_line},{crash_line}", "--porcelain", "--", file_rel],
                repo_root=repo_root,
                label="blame",
                warnings=warnings,
                per_cmd_timeout_sec=per_cmd_timeout_sec,
                deadline=deadline,
            )
            if blame_text is not None:
                blame_meta = _parse_blame_porcelain(blame_text, crash_line)
                if blame_meta is not None:
                    crash_file["blame"] = blame_meta

            diffstat_text = _run_git_budgeted(
                ["show", "--numstat", "--format=", "--", file_rel],
                repo_root=repo_root,
                label="crash file diffstat",
                warnings=warnings,
                per_cmd_timeout_sec=per_cmd_timeout_sec,
                deadline=deadline,
            )
            if diffstat_text is not None:
                context["crash_file_diffstat"] = _parse_numstat(diffstat_text)

        context["crash_file"] = crash_file

    log_text = _run_git_budgeted(
        [
            "log",
            "-n",
            str(max(1, max_commits)),
            "--date=iso-strict",
            f"--pretty=format:{_LOG_SEPARATOR}%n%h%n%H%n%an%n%ad%n%s",
            "--shortstat",
        ],
        repo_root=repo_root,
        label="recent commits",
        warnings=warnings,
        per_cmd_timeout_sec=per_cmd_timeout_sec,
        deadline=deadline,
    )
    if log_text is not None:
        recent = _parse_recent_commits(log_text)
        if recent:
            context["recent_commits"] = recent

    if warnings:
        context["_git_warnings"] = warnings
    return context


def get_git_context(*, dirty_mode: GitDirtyMode = "always") -> dict[str, Any] | None:
    """Capture git repository context.

    Returns:
        Dict with commit, branch, dirty status, or None if not in a git repo.

    Example output:
        {
            "commit": "abc123",
            "commit_full": "abc123def456...",
            "branch": "main",
            "dirty": True
        }
    """
    if dirty_mode not in {"always", "cached", "off"}:
        raise ValueError("dirty_mode must be 'always', 'cached', or 'off'")

    cached_hit, cached_context = _load_cached_git_context(dirty_mode=dirty_mode)
    if cached_hit:
        return cached_context

    repo_root_result = _run_git_command(["rev-parse", "--show-toplevel"])
    if isinstance(repo_root_result, _GitError):
        # Not a git repo, or git not available
        _store_git_context_cache(None, repo_root=None)
        return None
    repo_root: str = repo_root_result  # type: ignore[assignment]

    context: dict[str, Any] = {}
    _warnings: list[str] = []

    # Get commit hash (short and full)
    commit_short = _run_git_command(["rev-parse", "--short", "HEAD"], repo_root=repo_root)
    if isinstance(commit_short, str):
        context["commit"] = commit_short
    else:
        _warnings.append(f"commit hash: {commit_short.reason}")

    commit_full = _run_git_command(["rev-parse", "HEAD"], repo_root=repo_root)
    if isinstance(commit_full, str):
        context["commit_full"] = commit_full

    # Get branch name
    branch = _run_git_command(["rev-parse", "--abbrev-ref", "HEAD"], repo_root=repo_root)
    if isinstance(branch, str):
        context["branch"] = branch
        # Check for detached HEAD
        if branch == "HEAD":
            context["detached"] = True

    if dirty_mode != "off":
        dirty = _detect_dirty_status(repo_root)
        if dirty is None:
            status = _run_git_command(["status", "--porcelain"], repo_root=repo_root)
            if isinstance(status, str):
                dirty = len(status) > 0
        if dirty is not None:
            context["dirty"] = dirty

    if _warnings:
        context["_git_warnings"] = _warnings

    result = context if context else None
    _store_git_context_cache(result, repo_root=repo_root)
    return result



from __future__ import annotations

import shutil
from pathlib import Path
from typing import Any

from evals import run_eval
from evals.patchers.openai_compat_patcher import OpenAICompatPatcher
from llmdebug import _json as json

from ._openai_stub_server import OpenAIStubServer


def _make_case(tmp_path: Path) -> run_eval.CaseSpec:
    case_dir = tmp_path / "case_src"
    case_dir.mkdir(parents=True, exist_ok=True)
    (case_dir / "buggy.py").write_text(
        "import helper\n\n\ndef f() -> int:\n    return helper.value()\n",
        encoding="utf-8",
    )
    (case_dir / "helper.py").write_text("def value() -> int:\n    return 1\n", encoding="utf-8")
    (case_dir / "test_buggy.py").write_text(
        "from buggy import f\n\n\ndef test_f() -> None:\n    assert f() == 2\n",
        encoding="utf-8",
    )
    return run_eval.CaseSpec(
        case_id="openai_integration_case",
        path=case_dir,
        category="traceback_controls",
        difficulty="easy",
        snapshot_dependency="none",
        expected_exception="AssertionError",
        tags=[],
        python_args=["-m", "pytest", "-q"],
        timeout_sec=30,
    )


def _chat_json_response(content: str) -> dict[str, Any]:
    return {
        "choices": [{"message": {"content": content}}],
        "usage": {"prompt_tokens": 1, "completion_tokens": 1, "total_tokens": 2},
    }


def _diff_response_body() -> dict[str, Any]:
    diff = (
        "diff --git a/helper.py b/helper.py\n"
        "--- a/helper.py\n"
        "+++ b/helper.py\n"
        "@@ -1,2 +1,2 @@\n"
        " def value() -> int:\n"
        "-    return 1\n"
        "+    return 2\n"
    )
    return _chat_json_response(diff)


def test_openai_request_context_request_contract(tmp_path) -> None:
    with OpenAIStubServer() as server:
        server.enqueue_json(
            method="POST",
            path="/v1/chat/completions",
            body=_chat_json_response('{"need_more_context":false,"requests":[]}'),
        )
        patcher = OpenAICompatPatcher(
            base_url=server.base_url,
            model="demo-model",
            api_key="local",
            temperature=0.2,
            max_tokens=321,
        )

        result = patcher.request_context(
            prompt="Need more context?", context_dir=tmp_path, cwd=tmp_path
        )

        assert result is not None
        assert result.request_json is not None
        assert result.request_json["need_more_context"] is False

        req = server.last_request(method="POST", path="/v1/chat/completions")
        assert req is not None
        assert req["headers"]["authorization"] == "Bearer local"
        payload = req["json"]
        assert isinstance(payload, dict)
        assert payload["model"] == "demo-model"
        assert payload["temperature"] == 0.2
        assert payload["max_tokens"] == 321
        assert payload["response_format"]["type"] == "json_object"
        messages = payload["messages"]
        assert isinstance(messages, list)
        assert messages[0]["role"] == "system"
        assert messages[1]["role"] == "user"


def test_openai_compact_context_request_contract(tmp_path) -> None:
    with OpenAIStubServer() as server:
        server.enqueue_json(
            method="POST",
            path="/v1/chat/completions",
            body=_chat_json_response('{"compacted_context":"key traceback facts"}'),
        )
        patcher = OpenAICompatPatcher(
            base_url=server.base_url,
            model="demo-model",
            api_key="local",
            temperature=0.0,
            max_tokens=222,
        )

        result = patcher.compact_context(
            prompt="compact evidence",
            context_text="stdout ...\nsnapshot ...",
            context_dir=tmp_path,
            cwd=tmp_path,
            target_chars=32,
        )

        assert result is not None
        assert result.compacted_text == "key traceback facts"

        req = server.last_request(method="POST", path="/v1/chat/completions")
        assert req is not None
        payload = req["json"]
        assert isinstance(payload, dict)
        assert payload["model"] == "demo-model"
        assert payload["temperature"] == 0.0
        assert payload["max_tokens"] == 222
        assert payload["response_format"]["type"] == "json_object"
        messages = payload["messages"]
        assert isinstance(messages, list)
        assert messages[0]["role"] == "system"
        assert "<= 32 characters" in str(messages[0]["content"])
        assert messages[1]["role"] == "user"
        assert "Context to compact:" in str(messages[1]["content"])


def test_openai_edits_mode_response_format_fallback_contract(tmp_path) -> None:
    (tmp_path / "helper.py").write_text("def value() -> int:\n    return 1\n", encoding="utf-8")
    with OpenAIStubServer() as server:
        server.enqueue_json(
            method="POST",
            path="/v1/chat/completions",
            status=400,
            body={"error": {"message": "response_format not supported"}},
        )
        server.enqueue_json(
            method="POST",
            path="/v1/chat/completions",
            body=_chat_json_response(
                json.dumps(
                    {
                        "edits": [
                            {
                                "path": "helper.py",
                                "find": "return 1",
                                "replace": "return 2",
                                "count": 1,
                            }
                        ]
                    }
                )
            ),
        )
        patcher = OpenAICompatPatcher(
            base_url=server.base_url,
            model="demo-model",
            api_key="local",
            response_mode="edits",
        )

        result = patcher.propose_patch(prompt="fix helper", context_dir=tmp_path, cwd=tmp_path)

        assert result.diff is not None
        assert "diff --git a/helper.py b/helper.py" in result.diff

        reqs = server.requests_for(method="POST", path="/v1/chat/completions")
        assert len(reqs) == 2
        first_payload = reqs[0]["json"]
        second_payload = reqs[1]["json"]
        assert isinstance(first_payload, dict)
        assert isinstance(second_payload, dict)
        assert first_payload["response_format"]["type"] == "json_schema"
        assert second_payload["response_format"]["type"] == "json_object"


def test_openai_diff_mode_request_has_no_response_format(tmp_path) -> None:
    with OpenAIStubServer() as server:
        server.enqueue_json(method="POST", path="/v1/chat/completions", body=_diff_response_body())
        patcher = OpenAICompatPatcher(
            base_url=server.base_url,
            model="demo-model",
            api_key="local",
            response_mode="diff",
        )

        result = patcher.propose_patch(
            prompt="return unified diff", context_dir=tmp_path, cwd=tmp_path
        )

        assert result.diff is not None
        req = server.last_request(method="POST", path="/v1/chat/completions")
        assert req is not None
        payload = req["json"]
        assert isinstance(payload, dict)
        assert "response_format" not in payload


def test_openai_probe_context_window_endpoint_order(tmp_path) -> None:
    del tmp_path
    with OpenAIStubServer() as server:
        server.enqueue_json(
            method="GET",
            path="/api/v1/models",
            status=404,
            body={"error": {"message": "not found"}},
        )
        server.enqueue_json(
            method="GET",
            path="/api/v0/models",
            body={"data": [{"id": "demo-model", "max_context_length": 16384}]},
        )
        patcher = OpenAICompatPatcher(
            base_url=server.base_url_v1,
            model="demo-model",
            api_key="local",
        )

        tokens, meta = patcher.probe_context_window_tokens()

        assert tokens == 16384
        assert meta["context_window_probe_provider"] == "lmstudio_api_v0"
        assert meta["context_window_source"] == "max_context_length"
        request_paths = [item["path"] for item in server.requests if item["method"] == "GET"]
        assert request_paths == ["/api/v1/models", "/api/v0/models"]


def test_openai_probe_context_window_falls_back_to_openai_models_endpoint(tmp_path) -> None:
    del tmp_path
    with OpenAIStubServer() as server:
        server.enqueue_json(
            method="GET",
            path="/api/v1/models",
            status=404,
            body={"error": {"message": "not found"}},
        )
        server.enqueue_json(
            method="GET",
            path="/api/v0/models",
            status=404,
            body={"error": {"message": "not found"}},
        )
        server.enqueue_json(
            method="GET",
            path="/v1/models",
            body={"data": [{"id": "demo-model", "context_length": 8192}]},
        )
        patcher = OpenAICompatPatcher(
            base_url=server.base_url_v1,
            model="demo-model",
            api_key="local",
        )

        tokens, meta = patcher.probe_context_window_tokens()

        assert tokens == 8192
        assert meta["context_window_probe_provider"] == "openai_models"
        assert meta["context_window_source"] == "context_length"
        request_paths = [item["path"] for item in server.requests if item["method"] == "GET"]
        assert request_paths == ["/api/v1/models", "/api/v0/models", "/v1/models"]


def test_openai_edits_rejects_path_traversal_without_mutating_parent_file(tmp_path) -> None:
    outside_file = tmp_path.parent / "outside_sensitive_file.py"
    outside_file.write_text("FLAG = 'original'\n", encoding="utf-8")
    with OpenAIStubServer() as server:
        server.enqueue_json(
            method="POST",
            path="/v1/chat/completions",
            body=_chat_json_response(
                json.dumps(
                    {
                        "edits": [
                            {
                                "path": "../outside_sensitive_file.py",
                                "find": "original",
                                "replace": "modified",
                                "count": 1,
                            }
                        ]
                    }
                )
            ),
        )
        patcher = OpenAICompatPatcher(
            base_url=server.base_url_v1,
            model="demo-model",
            api_key="local",
            response_mode="edits",
        )

        result = patcher.propose_patch(prompt="fix", context_dir=tmp_path, cwd=tmp_path)

        assert result.diff is None
        assert result.meta is not None
        assert result.meta["error"] == "path_outside_case"
        assert outside_file.read_text(encoding="utf-8") == "FLAG = 'original'\n"


def test_openai_edits_rejects_directory_target(tmp_path) -> None:
    (tmp_path / "target_dir").mkdir(parents=True, exist_ok=True)
    with OpenAIStubServer() as server:
        server.enqueue_json(
            method="POST",
            path="/v1/chat/completions",
            body=_chat_json_response(
                json.dumps(
                    {
                        "edits": [
                            {
                                "path": "target_dir",
                                "find": "x",
                                "replace": "y",
                                "count": 1,
                            }
                        ]
                    }
                )
            ),
        )
        patcher = OpenAICompatPatcher(
            base_url=server.base_url_v1,
            model="demo-model",
            api_key="local",
            response_mode="edits",
        )

        result = patcher.propose_patch(prompt="fix", context_dir=tmp_path, cwd=tmp_path)

        assert result.diff is None
        assert result.meta is not None
        assert result.meta["error"] == "path_not_file"
        assert result.meta["path"] == "target_dir"


def test_openai_edits_surfaces_file_read_error(tmp_path, monkeypatch) -> None:
    target_file = tmp_path / "helper.py"
    target_file.write_text("def value() -> int:\n    return 1\n", encoding="utf-8")
    original_read_text = Path.read_text

    def _raising_read_text(self: Path, *args, **kwargs) -> str:
        if self.resolve() == target_file.resolve():
            raise OSError("simulated read failure")
        return original_read_text(self, *args, **kwargs)

    monkeypatch.setattr(Path, "read_text", _raising_read_text)
    with OpenAIStubServer() as server:
        server.enqueue_json(
            method="POST",
            path="/v1/chat/completions",
            body=_chat_json_response(
                json.dumps(
                    {
                        "edits": [
                            {
                                "path": "helper.py",
                                "find": "return 1",
                                "replace": "return 2",
                                "count": 1,
                            }
                        ]
                    }
                )
            ),
        )
        patcher = OpenAICompatPatcher(
            base_url=server.base_url_v1,
            model="demo-model",
            api_key="local",
            response_mode="edits",
        )

        result = patcher.propose_patch(prompt="fix", context_dir=tmp_path, cwd=tmp_path)

        assert result.diff is None
        assert result.meta is not None
        assert result.meta["error"] == "file_read_error"
        assert result.meta["path"] == "helper.py"
        assert "simulated read failure" in str(result.meta.get("exc"))


def test_stage_a_outside_file_request_uses_fallback_and_does_not_leak_secret(
    tmp_path, monkeypatch
) -> None:
    case = _make_case(tmp_path)
    secret_marker = "TOP_SECRET_SENTINEL_4731"
    outside_secret_path = tmp_path / "secret.txt"
    outside_secret_path.write_text(secret_marker, encoding="utf-8")
    workdir = tmp_path / "workdir_openai_stage_a"
    context_dir = tmp_path / "context_openai_stage_a"
    shutil.copytree(case.path, workdir)

    with OpenAIStubServer() as server:
        server.enqueue_json(
            method="POST",
            path="/v1/chat/completions",
            body=_chat_json_response(
                json.dumps(
                    {
                        "need_more_context": True,
                        "requests": [{"type": "file", "path": "../secret.txt"}],
                        "sufficiency_rationale": "need more context",
                    }
                )
            ),
        )
        server.enqueue_json(method="POST", path="/v1/chat/completions", body=_diff_response_body())
        patcher = OpenAICompatPatcher(
            base_url=server.base_url_v1,
            model="demo-model",
            api_key="local",
            response_mode="diff",
            force_patch_attempt=False,
        )

        call_counter = {"count": 0}

        def _fake_run_pytest(*args, **kwargs) -> run_eval.RunResult:
            del args, kwargs
            call_counter["count"] += 1
            if call_counter["count"] == 1:
                return run_eval.RunResult(
                    returncode=1,
                    duration_sec=0.01,
                    stdout="test_buggy.py::test_f\nE   AssertionError: expected 2 got 1\n",
                    stderr="",
                )
            return run_eval.RunResult(
                returncode=0,
                duration_sec=0.01,
                stdout="1 passed\n",
                stderr="",
            )

        monkeypatch.setattr(run_eval, "run_pytest", _fake_run_pytest)
        monkeypatch.setattr(
            run_eval,
            "read_latest_snapshot",
            lambda *_args, **_kwargs: (None, "missing snapshot"),
        )

        records = run_eval.run_one_condition(
            case=case,
            condition=run_eval.Condition.WITH_SNAPSHOT_STRUCTURED,
            workdir=workdir,
            context_dir_base=context_dir,
            patcher=patcher,
            k=1,
            output_format="json_compact",
            run_id="run-openai-stage-a-fallback",
            include_files_context=False,
            force_patch_attempt=False,
            context_protocol="staged",
        )

        assert len(records) == 1
        assert records[0]["success"] is True
        assert call_counter["count"] == 2
        assert records[0]["patch_meta"]["stage_a_used_fallback"] is True
        assert records[0]["patch_meta"]["stage_a_reason"] == "no_valid_stage_a_requests"

        requests = server.requests_for(method="POST", path="/v1/chat/completions")
        assert len(requests) == 2
        stage_b_payload = requests[1]["json"]
        assert isinstance(stage_b_payload, dict)
        messages = stage_b_payload["messages"]
        assert isinstance(messages, list)
        stage_b_user_prompt = str(messages[1]["content"])
        assert secret_marker not in stage_b_user_prompt
        assert "../secret.txt" not in stage_b_user_prompt



from __future__ import annotations

import sys
import types

import pytest

from evals.swebench.generate import _load_dataset_rows


def _install_fake_datasets(monkeypatch, rows) -> None:
    fake_datasets = types.ModuleType("datasets")
    fake_datasets.load_dataset = lambda _dataset_name, split: rows[split]
    monkeypatch.setitem(sys.modules, "datasets", fake_datasets)


def test_load_dataset_rows_returns_unique_instance_rows(monkeypatch) -> None:
    _install_fake_datasets(
        monkeypatch,
        {
            "test": [
                {"instance_id": "i-1", "repo": "demo/a"},
                {"instance_id": "i-2", "repo": "demo/b"},
            ]
        },
    )

    rows = _load_dataset_rows(dataset_name="demo", split="test")

    assert sorted(rows) == ["i-1", "i-2"]
    assert rows["i-1"]["repo"] == "demo/a"
    assert rows["i-2"]["repo"] == "demo/b"


def test_load_dataset_rows_rejects_duplicate_instance_ids(monkeypatch) -> None:
    _install_fake_datasets(
        monkeypatch,
        {
            "test": [
                {"instance_id": "dupe-1", "repo": "demo/a"},
                {"instance_id": "dupe-1", "repo": "demo/b"},
            ]
        },
    )

    with pytest.raises(RuntimeError, match="Duplicate instance_id values found"):
        _load_dataset_rows(dataset_name="demo", split="test")



"""RCA state tracking facade for MCP/debug workflows.

This module preserves the public RCA API while delegating internals to focused
identity/transition/backend helpers.
"""

from __future__ import annotations

import uuid
from datetime import datetime, timezone
from typing import Any, Literal

from . import rca_identity as _rca_identity
from .rca_backend import (
    get_rca_history as _get_rca_history_backend,
)
from .rca_backend import (
    get_rca_status as _get_rca_status_backend,
)
from .rca_backend import (
    persist_rca_record,
)
from .rca_engine import build_rca_attempt_record
from .rca_transitions import (
    DEFAULT_STATE,
)

RCABackend = Literal["sqlite", "jsonl"]


def _now_utc() -> str:
    return datetime.now(timezone.utc).replace(microsecond=0).isoformat().replace("+00:00", "Z")


def _new_attempt_id() -> str:
    return f"{int(datetime.now(timezone.utc).timestamp() * 1000)}-{uuid.uuid4().hex[:8]}"


def build_failure_signature(
    snapshot: dict[str, Any] | None,
    *,
    stdout: str = "",
    stderr: str = "",
) -> str:
    """Build a stable failure signature from snapshot (preferred) or textual output."""
    return _rca_identity.build_failure_signature(snapshot, stdout=stdout, stderr=stderr)


def derive_session_id(
    snapshot: dict[str, Any] | None,
    *,
    out_dir: str,
    fallback: str = "global",
) -> str:
    """Derive stable session id from out_dir + pytest nodeid/name when available."""
    return _rca_identity.derive_session_id(snapshot, out_dir=out_dir, fallback=fallback)


def get_rca_history(
    *,
    out_dir: str,
    session_id: str | None = None,
    limit: int = 20,
) -> list[dict[str, Any]]:
    """Load RCA history, newest-first."""
    return _get_rca_history_backend(out_dir=out_dir, session_id=session_id, limit=limit)


def get_rca_status(*, out_dir: str, session_id: str | None = None) -> dict[str, Any] | None:
    """Return latest RCA attempt for a session."""
    return _get_rca_status_backend(out_dir=out_dir, session_id=session_id)


def advance_rca_attempt(
    *,
    out_dir: str,
    session_id: str,
    action: str,
    failure_signature: str,
    evidence_refs: list[str] | None = None,
    hypothesis: str | None = None,
    experiment_plan: str | None = None,
    proposed_change_summary: str | None = None,
    verification_result: dict[str, Any] | None = None,
    gate_mode: str | None = None,
    exploratory: bool = False,
    override_reason: str | None = None,
    stall_threshold: int = 3,
    exploration_budget: int = 2,
) -> dict[str, Any]:
    """Advance RCA state machine and persist attempt row."""
    now = _now_utc()
    previous = get_rca_status(out_dir=out_dir, session_id=session_id)
    record = build_rca_attempt_record(
        previous=previous,
        attempt_id=_new_attempt_id(),
        session_id=session_id,
        action=action,
        failure_signature=failure_signature,
        evidence_refs=evidence_refs,
        hypothesis=hypothesis,
        experiment_plan=experiment_plan,
        proposed_change_summary=proposed_change_summary,
        verification_result=verification_result,
        gate_mode=gate_mode,
        exploratory=exploratory,
        override_reason=override_reason,
        stall_threshold=stall_threshold,
        exploration_budget=exploration_budget,
        now=now,
    )
    persist_rca_record(out_dir=out_dir, record=record)
    return record


# Backward-compatible alias for callers/tests that use this constant.
_DEFAULT_STATE = DEFAULT_STATE



from __future__ import annotations

import argparse
import csv
import glob
from pathlib import Path
from typing import Any

from . import _json as json


def main(argv: list[str] | None = None) -> int:
    parser = argparse.ArgumentParser(
        description=(
            "Export adaptive online-policy JSONL event logs to a flat CSV dataset "
            "for offline model training."
        )
    )
    parser.add_argument(
        "--events",
        action="append",
        required=True,
        help="Path or glob to adaptive policy event JSONL files (repeatable).",
    )
    parser.add_argument(
        "--output-csv",
        default="evals/results/adaptive_policy_dataset.csv",
        help="Output CSV path (default: evals/results/adaptive_policy_dataset.csv).",
    )
    parser.add_argument(
        "--min-propensity",
        type=float,
        default=1e-6,
        help="Lower-bound for propensity before inverse weighting (default: 1e-6).",
    )
    parser.add_argument(
        "--strict-input",
        action="store_true",
        help="Fail on first unreadable file or malformed JSON line.",
    )
    args = parser.parse_args(argv)

    min_propensity = max(1e-12, float(args.min_propensity))
    expanded = _expand_paths(args.events)
    if not expanded:
        print("No input matched --events.")
        return 1

    try:
        events, input_errors = _load_events(expanded, strict_input=bool(args.strict_input))
    except ValueError as exc:
        print(f"Input error: {exc}")
        return 2

    if not events:
        print("No valid policy events found.")
        _print_input_errors(input_errors)
        return 1

    rows = _events_to_rows(events=events, min_propensity=min_propensity)
    output_path = Path(str(args.output_csv))
    output_path.parent.mkdir(parents=True, exist_ok=True)
    _write_rows(rows=rows, output_path=output_path)

    print(f"Exported {len(rows)} policy events to {output_path}")
    _print_input_errors(input_errors)
    return 0


def _expand_paths(items: list[str]) -> list[str]:
    out: list[str] = []
    for item in items:
        matches = sorted(glob.glob(str(item)))
        out.extend(matches if matches else [str(item)])
    return out


def _load_events(
    paths: list[str], *, strict_input: bool
) -> tuple[list[dict[str, Any]], dict[str, int]]:
    input_errors = {
        "files_unreadable": 0,
        "json_lines_invalid": 0,
        "records_skipped": 0,
        "duplicate_events_seen": 0,
        "duplicate_events_resolved": 0,
    }
    deduped: dict[tuple[str, str, str, int, str], dict[str, Any]] = {}
    for src in paths:
        src_path = Path(src)
        try:
            lines = src_path.read_text(encoding="utf-8").splitlines()
        except OSError as exc:
            input_errors["files_unreadable"] += 1
            if strict_input:
                raise ValueError(f"{src}: unable to read ({exc})") from exc
            continue
        for line_no, raw_line in enumerate(lines, start=1):
            line = raw_line.strip()
            if not line:
                continue
            try:
                payload = json.loads(line)
            except json.JSONDecodeError as exc:
                input_errors["json_lines_invalid"] += 1
                if strict_input:
                    raise ValueError(f"{src}:{line_no}: invalid JSON ({exc})") from exc
                continue
            if not isinstance(payload, dict):
                input_errors["records_skipped"] += 1
                if strict_input:
                    raise ValueError(f"{src}:{line_no}: JSON root is not an object")
                continue
            key = _event_key(payload)
            if key is None:
                input_errors["records_skipped"] += 1
                if strict_input:
                    raise ValueError(f"{src}:{line_no}: missing run/case/condition/attempt")
                continue
            if key in deduped:
                input_errors["duplicate_events_seen"] += 1
            deduped[key] = payload
    input_errors["duplicate_events_resolved"] = input_errors["duplicate_events_seen"]
    ordered = [deduped[key] for key in sorted(deduped)]
    return ordered, input_errors


def _event_key(payload: dict[str, Any]) -> tuple[str, str, str, int, str] | None:
    run_id = str(payload.get("run_id") or "").strip()
    case_id = str(payload.get("case_id") or "").strip()
    condition = str(payload.get("condition") or "").strip()
    attempt = _safe_int(payload.get("attempt"), default=-1)
    if not run_id or not case_id or not condition or attempt < 0:
        return None
    decision_id = str(payload.get("decision_id") or "").strip()
    if not decision_id:
        decision_id = str(payload.get("created_at_utc") or "").strip()
    if not decision_id:
        decision_id = "unknown"
    return run_id, case_id, condition, attempt, decision_id


def _events_to_rows(*, events: list[dict[str, Any]], min_propensity: float) -> list[dict[str, Any]]:
    feature_names = sorted(
        {
            str(name)
            for event in events
            for name in (
                event.get("feature_values", {}).keys()
                if isinstance(event.get("feature_values"), dict)
                else []
            )
        }
    )
    rows: list[dict[str, Any]] = []
    for event in events:
        propensity = _safe_float(event.get("propensity"), default=1.0)
        propensity_clipped = max(min_propensity, propensity)
        scores = event.get("scores")
        scores_dict = scores if isinstance(scores, dict) else {}
        features = event.get("feature_values")
        feature_dict = features if isinstance(features, dict) else {}
        row: dict[str, Any] = {
            "run_id": str(event.get("run_id") or ""),
            "case_id": str(event.get("case_id") or ""),
            "condition": str(event.get("condition") or ""),
            "attempt": _safe_int(event.get("attempt"), default=0),
            "note": str(event.get("note") or ""),
            "decision_id": str(event.get("decision_id") or ""),
            "decision_index": _safe_int(event.get("decision_index"), default=0),
            "mode": str(event.get("mode") or ""),
            "algorithm": str(event.get("algorithm") or ""),
            "decision_strategy": str(event.get("decision_strategy") or ""),
            "selected_action": str(event.get("selected_action") or ""),
            "applied_action": str(event.get("applied_action") or ""),
            "default_action": str(event.get("default_action") or ""),
            "override_reason": str(event.get("override_reason") or ""),
            "propensity": propensity,
            "propensity_clipped": propensity_clipped,
            "importance_weight": (1.0 / propensity_clipped),
            "reward": _safe_float(event.get("reward"), default=0.0),
            "reward_success": _safe_float(event.get("reward_success"), default=0.0),
            "reward_token_penalty": _safe_float(event.get("reward_token_penalty"), default=0.0),
            "reward_latency_penalty": _safe_float(
                event.get("reward_latency_penalty"), default=0.0
            ),
            "success": 1 if bool(event.get("success")) else 0,
            "tokens_total": _safe_int(event.get("tokens_total"), default=0),
            "attempt_wall_sec": _safe_float(event.get("attempt_wall_sec"), default=0.0),
            "model_updated": 1 if bool(event.get("model_updated")) else 0,
            "update_reason": str(event.get("update_reason") or ""),
            "score_skip_helper": _safe_float(scores_dict.get("skip_helper"), default=0.0),
            "score_invoke_helper": _safe_float(scores_dict.get("invoke_helper"), default=0.0),
            "created_at_utc": str(event.get("created_at_utc") or ""),
        }
        for feature_name in feature_names:
            row[f"feature_{feature_name}"] = _safe_float(feature_dict.get(feature_name), default=0.0)
        rows.append(row)
    return rows


def _write_rows(*, rows: list[dict[str, Any]], output_path: Path) -> None:
    fieldnames = sorted({key for row in rows for key in row})
    with output_path.open("w", encoding="utf-8", newline="") as handle:
        writer = csv.DictWriter(handle, fieldnames=fieldnames)
        writer.writeheader()
        writer.writerows(rows)


def _print_input_errors(input_errors: dict[str, int]) -> None:
    files_unreadable = int(input_errors.get("files_unreadable", 0))
    json_lines_invalid = int(input_errors.get("json_lines_invalid", 0))
    records_skipped = int(input_errors.get("records_skipped", 0))
    duplicate_events_seen = int(input_errors.get("duplicate_events_seen", 0))
    duplicate_events_resolved = int(input_errors.get("duplicate_events_resolved", 0))
    if (
        files_unreadable == 0
        and json_lines_invalid == 0
        and records_skipped == 0
        and duplicate_events_seen == 0
    ):
        return
    print("\nInput diagnostics:")
    if files_unreadable > 0 or json_lines_invalid > 0 or records_skipped > 0:
        print(
            "- errors: "
            f"files_unreadable={files_unreadable}, "
            f"json_lines_invalid={json_lines_invalid}, "
            f"records_skipped={records_skipped}"
        )
    if duplicate_events_seen > 0:
        print(
            "- dedup: "
            f"duplicate_events_seen={duplicate_events_seen}, "
            f"duplicate_events_resolved={duplicate_events_resolved}"
        )


def _safe_float(value: Any, *, default: float) -> float:
    try:
        out = float(value)
    except (TypeError, ValueError):
        return float(default)
    if out != out:  # NaN check
        return float(default)
    return out


def _safe_int(value: Any, *, default: int) -> int:
    try:
        return int(value)
    except (TypeError, ValueError):
        return int(default)


if __name__ == "__main__":
    raise SystemExit(main())
